Compiler.Control.Print.printDepth  :=  30;
Compiler.Control.Print.printLength := 100;
Compiler.Control.Print.stringDepth := 200;

exception InternalError
exception Unsatisfied

type var = string
type tag = int

(* extended attibutes *)
datatype attr =
    Az | Ac | Ao | Aco
  | Avar of var;

(* extended usages *)
datatype usage =
    Uzero
  | Uout  of attr  * usage
  | Uin   of attr  * usage
  | Upar  of usage * usage
  | Uglb  of usage * usage
  | Urec  of usage * usage 
        (* the first 'usage' must be of the form (Uvar s) *)
  | Urep  of usage                       (* *U *)
  | Uvar  of var
  | Uuof  of typ                         (* Uof(t) *)

(* extended types *)
and typ =
    Tbool
  | Tchan of (typ list) * tag * usage
  | Tvar of var
  | Tadd of typ * typ                    (* t1 + t2 *)
  | Tglb of typ * typ                    (* t1 |~| t2 *)
  | Trep of typ                          (* !t *)

  | Tif  of typ                          (* used in pt(if x then ...) *)

(* Process expression *)
datatype proc = 
    Pzero  
  | Pnu  of var  * proc
  | Pout of var  * attr * var list * proc
  | Pin  of var  * attr * var list * proc
  | Ppar of proc * proc 
  | Prep of proc                     (* !P *)
  | Pif  of var  * proc * proc;

type typing = (var * typ)
type env = typing list

(* constraint *)

(* on time tags *)
datatype cg =
    CGleq of tag * tag                   (* t1<= t2 *)

(* on attributes *)
datatype ca =
    CAinc of attr * attr                 (* a1 \subseteq a2 *)
    
(* on obligation of type/usage: o1,...,on *)
datatype co =
    COt of typ                           (* Ob(t) *)
  | COu of usage                         (* Ob(U) *)

(* on usages *)
datatype cu =
    CUleq of usage * usage               (* U1 <= U2 *)
  | CUnoob of usage                      (* noob(U) *)
  | CUobcap of usage list * ca           (* ob(U1)|...|ob(Un) -> ca *)
  | CUobct of usage * cg                 (* ob(U) -> cg *)
  | CUrel of usage                       (* rel(U) *)

(* on types *)
datatype ct =
    CTid of typ * typ                    (* t1 == t2 *)
  | CTsim of typ * typ                   (* t1 ~~ t2 *)
  | CTleq of typ * typ                   (* t1 <= t2 *)
  | CTnoob of typ                        (* noob(t) *)
  | CTrel of typ                         (* rel(t) *)
  | CTobcap of co list * ca              (* o1|...|on -> ca *)
  | CTtime of tag * typ                  (* tag R t *)

datatype con =
    Cfalse
  | CT of ct
  | CU of cu
  | CA of ca

(* replacement: used in constraint solving *)
datatype replace =
    RT of var * typ                      (* type replacement *)
  | RG of tag * tag                      (* time tag replacement *)
  | RU of usage * usage                  (* usage replacement *)

(* extended constraints *)

(* left/right hand side of extented constraint *)
datatype eclr =
    ECAt of tag * tag       (* g1 R g2 *)
  | ECAcap of var           (* Ac << (Avar s) *)
  | ECAob of var            (* Ao << (Avar s) *)
  | ECAobo of var           (* ob_o(Uvar s) *)
  | ECAobi of var           (* ob_i(Uvar s) *)
  | ECtrue                  (* true *)
  | ECfalse                 (* false *)
  | ECand of eclr list      (* c1 & c2 & ... *)
  | ECor of eclr list       (* c1 | c2 | ... *)
         
(* eclr1 -> eclr2 *)
datatype ec = EC of eclr * eclr

