(* 'pplist f str1 str2 []' returns 'str1' and
   'pplist f str1 str2 [fx1,...,xn]' returns '(f x1)^str2^...^(f xn)' *)
fun pplist _ str1 _ [] = str1
  | pplist f _ _ [u] = f u
  | pplist f str1 str2 (u::us) = (f u)^str2^(pplist f str1 str2 us)

(* time tag *)
fun ppg tg = "g" ^ Int.toString tg

(* attribute *)
fun ppa (Avar x) = x
  | ppa Az = ""
  | ppa Ao = "o"
  | ppa Ac = "c"
  | ppa Aco = "co";


(* if true, the usage/type needs "(",")" *)
fun compu (Uglb _) = true
  | compu (Upar _) = true
  | compu (Urec _) = true
  | compu _ = false

fun compt (Tadd _) = true
  | compt (Tglb _) = true
  | compt _ = false

(* usage, type, list of usages/types *)
fun ppu (Uglb(u,v)) =
    if compu u then
	if compu v then "("^(ppu u)^")^("^(ppu v)^")"
	else "("^(ppu u)^")^"^(ppu v)
    else if compu v then (ppu u)^"^("^(ppu v)^")"
	 else (ppu u)^"^"^(ppu v)
  | ppu (Uin(a,Uzero)) = "I"^(ppa a)
  | ppu (Uin(a,u)) =
	     "I"^(ppa a)^"."^(ppu u)
  | ppu (Uout(a,Uzero)) = "O"^(ppa a)
  | ppu (Uout(a,u)) =
	     "O"^(ppa a)^"."^(ppu u)
  | ppu (Upar(u,v)) = 
	    if compu u then
		if compu v then "("^(ppu u)^")||("^(ppu v)^")"
		else "("^(ppu u)^")||"^(ppu v)
	    else if compu v then (ppu u)^"||("^(ppu v)^")"
		 else (ppu u)^"||"^(ppu v)
  | ppu (Urec(u,Urec(v,w))) = "rec "^(ppu u)^"."^(ppu (Urec(v,w)))
  | ppu (Urec(u,v)) =
		     if compu v then "rec "^(ppu u)^".("^(ppu v)^")"
		     else "rec "^(ppu u)^"."^(ppu v)
  | ppu (Urep(u)) = if compu u then "*("^(ppu u)^")" else "*"^(ppu u)
  | ppu (Uuof(t)) = "Uof("^(ppt t)^")"
  | ppu (Uvar x) = x
  | ppu Uzero = "0"
and
    ppus str us = pplist ppu "" str us
and
    ppt (Tvar x) =x
  | ppt (Tbool) = "Bool"
  | ppt (Tchan(ts,g,u)) = "["^(ppts ts)^"]^"^(ppg g)^"/"^(ppu u)
  | ppt (Tadd(t1,t2)) =
    (case (compt t1, compt t2) of
	 (true,true)  => "("^(ppt t1)^")+("^(ppt t2)^")"
       | (true,false) => "("^(ppt t1)^")+"^(ppt t2)
       | (false,true) => (ppt t1)^"+("^(ppt t2)^")"
       | (false,false) => (ppt t1)^"+"^(ppt t2)
	     )
  | ppt (Trep(t)) = if compt t then "!("^(ppt t)^")"
		    else "!"^(ppt t)
  | ppt (Tglb(t1,t2)) = 
    (case (compt t1, compt t2) of
	 (true,true)  => "("^(ppt t1)^")^("^(ppt t2)^")"
       | (true,false) => "("^(ppt t1)^")^"^(ppt t2)
       | (false,true) => (ppt t1)^"^("^(ppt t2)^")"
       | (false,false) => (ppt t1)^"^"^(ppt t2)
	     )
  | ppt (Tif(t)) = "Tif(" ^ (ppt t) ^ ")"
and
    ppts ts = pplist ppt "" ", " ts
 
(* process *)
local
    fun ppa (Avar x) = x
      | ppa Az = ""
      | ppa Ao = "^o"
      | ppa Ac = "^c"
      | ppa Aco = "^co"
in
    fun ppp Pzero = "0"
      | ppp (Pnu(v,Ppar(p,q))) = "(new " ^ v ^ ")" ^ (ppp (Ppar(p,q)))
      | ppp (Pnu(v,Pnu(w,p))) = "(new " ^ v ^ ")" ^ (ppp (Pnu(w,p)))
      | ppp (Pnu(v,p)) = "(new " ^ v ^ ")(" ^ (ppp p) ^ ")"
      | ppp (Pin(x,a,vs,p)) =
	if p=Pzero then 
	    (x ^ "?" ^ (ppa a) ^ "[" ^ (pplist id "" "," vs) ^ "]")
	else
	    (x ^ "?" ^ (ppa a) ^ "[" ^ (pplist id "" "," vs) ^ "]." ^ (ppp p))
      | ppp (Pout(x,a,vs,p)) =
	    if p=Pzero then
		x ^ "!" ^ (ppa a) ^ "[" ^ (pplist id "" "," vs) ^ "]"
	    else
		x ^ "!" ^ (ppa a) ^ "[" ^ (pplist id "" "," vs) ^ "]." ^ (ppp p)
      | ppp (Ppar(p,q)) = ("("^(ppp p)^" | "^(ppp q)^")")
      | ppp (Prep(p)) = "*"^(ppp p)
      | ppp (Pif(x,p,q)) = ("if "^x^" then "^(ppp p)^" else "^(ppp q))
end

(*******************************************************************)

(* constraint on time tags *)
fun ppcg (CGleq(g1,g2)) = (ppg g1)^" < "^(ppg g2)

(* constraint[s] on attributes *) 
local 
    fun ppa (Avar x) = x
      | ppa Az = "0"
      | ppa Ao = "o"
      | ppa Ac = "c"
      | ppa Aco = "co"
in
    fun ppca (CAinc(a1,a2)) = (ppa a1)^"<<"^(ppa a2)
(*
      | ppca (CAleq(a1,a2)) = (ppa a1)^"<="^(ppa a2)
      | ppca (CAfalse) = "False"
*)
    and ppcas s cas = pplist ppca "" s cas
end

(* constraint[s] on obligation *)
fun ppco (COt t) = "ob("^(ppt t)^")"
  | ppco (COu u) = "ob("^(ppu u)^")"
and ppcos cos = pplist ppco "False" " | " cos 

(* constraint on usages *)
fun ppcu (CUleq(u1,u2)) = (ppu u1)^" <= "^(ppu u2)
  | ppcu (CUnoob u) = "noob("^(ppu u)^")"
  | ppcu (CUrel u) = "rel("^(ppu u)^")"
  | ppcu (CUobcap(us,ca)) = "ob("^(ppus " | " us)^") -> "^(ppca ca)
  | ppcu (CUobct(u,cg)) = "ob("^(ppu u)^") -> "^(ppcg cg)
    
(* constraint on types *)
fun ppct (CTsim(t1,t2)) = (ppt t1)^" ~~ "^(ppt t2)
  | ppct (CTid(t1,t2)) = (ppt t1)^" == "^(ppt t2)
  | ppct (CTleq(t1,t2)) = (ppt t1)^" <= "^(ppt t2)
  | ppct (CTnoob t) = "noob("^(ppt t)^")"
  | ppct (CTrel t) = "rel("^(ppt t)^")"
  | ppct (CTobcap(cos,ca)) = (ppcos cos)^" -> "^(ppca ca)
  | ppct (CTtime(tg,t)) = (ppg tg)^" < "^(ppt t)

(* constraint *)
fun ppcon Cfalse = "False"
  | ppcon (CT ct) = ppct ct
  | ppcon (CU cu) = ppcu cu
  | ppcon (CA ca) = ppca ca
(*  | ppcon (CG cg) = ppcg cg *)

(* constraints *)
fun ppcons s cs = pplist ppcon "" s cs

(* typing *)
fun ppe ((v,t):typing) = v^":"^(ppt t)
fun ppenv (es:env) = pplist ppe "" ", " es

(* principal typing *)
fun pppt (es,cs) = "({" ^ (ppenv es) ^ "},\n {\t"^(ppcons "\n\t" cs)^" })"

(* replacement *)
fun pprp (RT(v,t)) = "substitute "^(ppt t)^" for "^v
  | pprp (RU(u1,u2)) = "("^(ppu u1)^" -> "^(ppu u2)^")"
  | pprp _ = raise InternalError

(* substitution: a list of usage replacements *)
fun pptheta theta = pplist pprp "" ", " theta

(* extended constraints *)
local
    fun ppsub _ (ECAt(g1,g2)) = (ppg g1)^" < "^(ppg g2)
      | ppsub _ (ECAcap s) = "c<<"^s
      | ppsub _ (ECAob s) = "o<<"^s
      | ppsub _ (ECAobo s) = "ob_o("^s^")"
      | ppsub _ (ECAobi s) = "ob_i("^s^")"
      | ppsub _ ECtrue = "true"
      | ppsub _ ECfalse = "false"
      | ppsub _ (ECand [eclr]) = ppsub false eclr
      | ppsub paren (ECand ecs) =
	if paren
	    then "("^(pplist (ppsub true) "" " & " ecs)^")"
	else pplist (ppsub true) "" " & " ecs
      | ppsub _ (ECor [eclr]) = ppsub false eclr
      | ppsub paren (ECor ecs) =
	    if paren
		then "("^(pplist (ppsub true) "false" " | " ecs)^")"
	    else pplist (ppsub true) "false" " | " ecs
in
    val ppeclr = ppsub false
end

fun ppec (EC(ec1,ec2)) = (ppeclr ec1)^" -> "^(ppeclr ec2) 
fun ppecs ecs = pplist ppec "" "\n" ecs
