(* Rep(t) *)
fun rep (Tchan(ts,tg,u)) = Tchan(ts,tg,Uzero)
  | rep (Tadd(t1,t2)) = rep t1
  | rep (Tglb(t1,t2)) = rep t1
  | rep (Trep(t)) = rep t
  | rep (Tif(t)) = rep t
  | rep other = other            (* Tbool, Tvar *)

(* 'assocenv "x" env' finds 'env("x")' and returns 'env\{"x"}' *)
local
    fun sub _ es1 []  = (NONE,rev es1)
      | sub v1 es1 ((e as (v2,t))::es2) =
	if v1 = v2 
	    then (SOME t, (rev es1) @ es2) 
	else (sub v1 (e::es1) es2)
in 
    fun assocenv v (es:env) = sub v [] es
end

(* 'eremove env [v1,v2,...,vn]' returns env\{v1,v2,...,vn} *)
fun eremove ([]:env) vs = []
  | eremove ((e as (v,_))::es) vs =
    if member v vs then eremove es vs
	else (e::(eremove es vs))

(* makes an environment env1 + env2 + ... *)
fun addenv [] es2 = es2
  | addenv ((e as (v,t1))::es1) es2 = 
    (case assocenv v es2 of
	 (SOME t2,es2rest) => ((v, Tadd(t1,t2))::(addenv es1 es2rest))
       | (NONE,_) => (e::(addenv es1 es2))
	     )
fun addenvs envs = fold addenv envs

(* makes constraints env1 ~ env2 ~ ... *)
fun simenv [] es2 = []
  | simenv ((e as (v,t1))::es1) es2 = 
    (case assocenv v es2 of
	 (SOME t2,es2rest) => ((CT(CTsim(rep t1, rep t2)))::(simenv es1 es2rest))
       | (NONE,_) => simenv es1 es2
	     )
fun simenvs [] = raise Empty
  | simenvs [es] = []
  | simenvs (es::ess) = (List.concat (map (simenv es) ess)) @ (simenvs ess)

(* makes an environment env1 |~| env2 *)
fun glbenv [] es2 = map (fn (v,t) => (v,Tif(t))) es2
  | glbenv ((e as (v,t1))::es1) es2 = 
    (case assocenv v es2 of
	(SOME t2,es2rest) => ((v, Tglb(t1,t2))::(glbenv es1 es2rest))
      | (NONE,_) => (v, Tif(t1))::(glbenv es1 es2)
	    )

(* make a list, has one constraint 'ob(env) => Ac \subseteq a' *) 
fun makeobcap (es:env) a =
    [CT(CTobcap(map ((fn t => COt(t)) o snd) es, CAinc(Ac,a)))]
(* makes constraints 'tg R env' *)
fun makecttime tg (es:env) =
    map ((fn t => CT(CTtime(tg,t))) o snd) es
(* finds 'true' and 'false' in env and makes constraints on types *) 
fun makectid ([]:env) = []
  | makectid ((v,t)::es) = if v = "true" orelse v = "false"
			       then (CT(CTid(t,Tbool)) :: (makectid es))
			   else makectid es
(* used in 'pt x?[y1,...,yn].P' *)
fun makectleq ([]:env) es2 = []
  | makectleq ((v,t1)::es1) es2 =         (* 'e' is of the form '(y_i,rho_i)' *)
    (case assocenv v es2 of
	 (SOME t2,_) => (CT(CTleq(t1,t2)))::(CT(CTsim(t1,rep t2)))::(makectleq es1 es2)
       | (NONE,_) => (CT(CTnoob t1))::(makectleq es1 es2)
	     )

(* assign to the name 'v' a new type variable 'tn' *)
fun newtype v = (v,Tvar(new "t"))

(* pt(P) returns a pair '(es,cs)' where 'es' is a type environment and
     'cs' is a set of constraints. *)
fun pt Pzero = ([],[])
  | pt (Pout(x,a,vs,p)) =
    let	val (es1,cs1) = pt(p)
	val es_vs = (map newtype vs)
	val ts_vs = (map snd es_vs)
	val tg = newtag ()
	val a1 = Avar(new "a") 
	val (t_op,esrest) = assocenv x es1
	val (cs2,u) = 
	    (case t_op of
		 SOME t =>
		     ([CT(CTsim(Tchan(ts_vs, tg, Uvar(new "u")),
				rep t))], Uuof t)
	       | NONE =>([],Uzero))
	val es2 = addenvs (esrest :: (map sng es_vs))
            (* es2 = es1\{x} + v_1:rho_1 + ... + v_n:rho_n *) 
    in
	(
	 addenv [(x,Tchan(ts_vs, tg, Uout(a1,u)))] es2, 
	 cs1 @ cs2 @ [CA(CAinc(a,a1))] @ (makeobcap es2 a1)@
	 (simenvs ([(x,Tchan(ts_vs, tg, Uout(a1,u)))]::
		   esrest :: (map sng es_vs))) @
	 (makecttime tg es2) @ (makectid es_vs)
	 )
    end     (* pt(Pout...) *)

  | pt (Pin(x,a,vs,p)) =
    let	val (es1,cs1) = pt(p)
	val es_vs = (map newtype vs)
	val ts_vs = (map snd es_vs)
	val tg = newtag () 
	val a1 = Avar(new "a") 
	val (t_op,esrest) = assocenv x es1
	val (cs2,u) =
	    (case t_op of
		 SOME t => 
		     ([CT(CTsim(Tchan(ts_vs, tg, Uvar(new "u")),
				(*  Uzero is OK? ^^^^^^^^^^^^^ *)
				rep t))], Uuof(t))
	       | NONE => ([],Uzero))
	val esrest2 = eremove esrest vs
         (* esrest2 = es1\{x,y1,...,yn} *)
    in
	(
	 (x,Tchan(ts_vs, tg, Uin(a1,u))) :: esrest2, 
	 cs1 @ cs2 @ [CA(CAinc(a,a1))] @
	 (makeobcap esrest2 a1) @
	 (makectleq es_vs es1) @
	 (makecttime tg esrest2)
	 )
    end     (* pt(Pin...) *)

  | pt (Ppar(p1,p2)) =
    let
	val (es1,cs1) = pt(p1)
	val (es2,cs2) = pt(p2)
    in
	(
	 addenv es1 es2,
	 cs1 @ cs2 @ (simenv es1 es2)
	 )
    end
  | pt (Pnu(v,p)) =
    let val (es1,cs1) = pt(p) in
	(case assocenv v es1 of

	    (SOME t,esrest) => (esrest, cs1 @ [CT(CTrel(t))]) 
(*
	    (SOME t,esrest) => (es1, cs1 @ [CT(CTrel(t))]) 
*)
	  | (NONE, _) => (es1,cs1))
    end
  | pt (Pif(v,p1,p2)) =
    let val (es1,cs1) = pt(p1)
	val (es2,cs2) = pt(p2) in
	    (addenv (glbenv es1 es2) [(v,Tbool)],
	     cs1 @ cs2 @ (simenvs [es1, es2, [(v,Tbool)]]))
    end
  | pt (Prep(p)) =
    let val (es,cs) = pt(p) in
	(map (fn (v,t) => (v,Trep(t))) es, cs)
    end

fun pt2 p = (reset (); (println (pppt (pt p))))
