(* replace something in usage u0 *)
fun substu (u0, rp) =
    (case u0 of
	 Uout(a,u)  => Uout(a,substu(u,rp))
       | Uin(a,u)   => Uin(a,substu(u,rp))
       | Upar(u,v)  => Upar(substu(u,rp),substu(v,rp))
       | Uglb(u,v)  => Uglb(substu(u,rp),substu(v,rp))
       | Urec (u,v) => Urec(u,substu(v,rp))
       | Urep(u)    => Urep(substu(u,rp))
       | Uuof(t)    => Uuof(substt(t,rp))
       | Uzero      => Uzero
       | Uvar s     => 
	     (case rp of
		  RU(u1,u2) => if u0=u1 then u2 else u0
		| _         => u0))    (* substu *)

(* replace something in type t *)
and substt (t,rp) =
    (case rp of
	 RT(x,tp) =>
	     (case t of
		  Tvar(str) => if (x=str) then tp else t
		| Tadd(t1,t2) => Tadd(substt(t1,rp), substt(t2,rp))
		| Tglb(t1,t2) => Tglb(substt(t1,rp), substt(t2,rp))
		| Trep(t1) => Trep(substt(t1,rp))
		| Tchan(ts,tg,u) => Tchan(substts(ts,rp),tg,substu(u,rp))
		| Tif(t1) => Tif(substt(t1,rp))
		| _ => t)   (* Tbool *)
       | RG(g1,g2) =>
	     (case t of
		  Tadd(t1,t2) => Tadd(substt(t1,rp), substt(t2,rp))
		| Tglb(t1,t2) => Tglb(substt(t1,rp), substt(t2,rp))
		| Trep(t1) => Trep(substt(t1,rp))
		| Tchan(ts,tg,u) => 
		      if tg=g1 then
			  Tchan(substts(ts,rp),g2,substu(u,rp)) 
		      else Tchan(substts(ts,rp),tg,substu(u,rp))
		| Tif(t1) => Tif(substt(t1,rp))
		| _ => t)  (* Tbool, Tvar *)
       | RU(u1,u2) =>
	     (case t of
		  Tadd(t1,t2) => Tadd(substt(t1,rp), substt(t2,rp))
		| Tglb(t1,t2) => Tglb(substt(t1,rp), substt(t2,rp))
		| Trep(t1) => Trep(substt(t1,rp))
		| Tchan(ts,tg,u) => Tchan(substts(ts,rp),tg,substu(u,rp))
		| Tif(t1) => Tif(substt(t1,rp))
		| _ => t)  (* Tbool, Tvar *)
		  )   (* substt *)

(* in usages/types *)
and substus (us,rp) = map (fn u => substu(u,rp)) us 
and substts (ts,rp) = map (fn t => substt(t,rp)) ts

(* in o_n's*)
fun substcos ([],_) = []
  | substcos ((COt t)::os,rp) = COt(substt(t,rp))::(substcos(os,rp))
  | substcos ((COu u)::os,rp) = COu(substu(u,rp))::(substcos(os,rp))
    
(* in a time tag g *)
fun substg (g,RG(g1,g2)) = if g=g1 then g2 else g
  | substg (g,_) = g
    
(* in a constraint on types *)
fun substct (CTid(t1,t2),rp)     = CTid(substt(t1,rp),substt(t2,rp)) 
  | substct (CTsim(t1,t2),rp)    = CTsim(substt(t1,rp),substt(t2,rp)) 
  | substct (CTleq(t1,t2),rp)    = CTleq(substt(t1,rp),substt(t2,rp)) 
  | substct (CTnoob t1,rp)       = CTnoob(substt(t1,rp)) 
  | substct (CTrel t1,rp)        = CTrel(substt(t1,rp))
  | substct (CTobcap(obs,ca),rp) = CTobcap(substcos(obs,rp),ca)
  | substct (CTtime(g,t1),rp)    = CTtime(substg(g,rp),substt(t1,rp))

(* in a constraint on usages *)
fun substcu (CUleq(u1,u2),rp)    = CUleq(substu(u1,rp),substu(u2,rp))
  | substcu (CUnoob u1,rp)       = CUnoob(substu(u1,rp))
  | substcu (CUobcap(us, ca),rp) = CUobcap(substus(us,rp),ca)
  | substcu (CUobct(u1,cg),rp)   = CUobct(substu(u1,rp),cg)
  | substcu (CUrel u1,rp)        = CUrel(substu(u1,rp))

(* in constraints *)
fun substcs ([],_) = []
  | substcs ((CT ct)::cs,rp) = CT(substct(ct,rp))::(substcs(cs,rp))
  | substcs ((CU cu)::cs,rp) = CU(substcu(cu,rp))::(substcs(cs,rp))
  | substcs (con::cs,rp) = con::(substcs(cs,rp))   (* CA,CG and Cfalse *)

local
    (* simplifies types/usages in a constraint *)
    fun smplu (Uout(a,u)) = Uout(a,smplu u)
      | smplu (Uin(a,u)) = Uin(a,smplu u)
      | smplu (Upar(u,v)) = Upar(smplu u,smplu v)
      | smplu (Uglb(u,v)) = Uglb(smplu u,smplu v)
      | smplu (Urec(u,v)) = Urec(u,smplu v)
      | smplu (Urep u) = Urep (smplu u)
      | smplu (Uuof t) =
	(case smplt t of
	     Tchan(_,_,u) => u
	   | _ => Uuof(t))
      | smplu u = u   (* Uzero, Uvar *)
	     
    and smplt (Tadd(t1,t2)) = 
	(case (smplt t1, smplt t2) of
	     (Tbool, Tbool) => Tbool
	   | (Tchan(ts1,g1,u1),Tchan(ts2,g2,u2)) =>
		 if (ts1=ts2) andalso (g1=g2) then 
		     Tchan(ts1,g1,Upar(u1,u2))
		 else
		     Tadd(t1,t2)
	   | _ => Tadd(t1,t2))
      | smplt (Tglb(t1,t2)) =
	     (case (smplt t1, smplt t2) of
		  (Tbool, Tbool) => Tbool
		| (Tchan(ts1,g1,u1),Tchan(ts2,g2,u2)) =>
		      if (ts1=ts2) andalso (g1=g2) then 
			  Tchan(ts1,g1,Uglb(u1,u2))
		      else
			  Tglb(t1,t2)
		| _ => Tglb(t1,t2))
      | smplt (Trep t) =
		  (case smplt t of
		       Tbool => Tbool
		     | Tchan(ts,g,u) => Tchan(ts,g,Urep u)
		     | _ => Trep t)
      | smplt (Tif(t)) =
		       (case smplt t of
			    Tbool => Tbool
			  | Tchan(ts,g,u) => Tchan(ts,g,Uglb(u,Uzero))
			  | _ => Tif(t))
      | smplt (Tchan(ts,g,u)) = Tchan(map smplt ts,g,smplu u)
      | smplt t = t   (* Tbool, Tvar *)
			    
    fun smplcos [] = []
      | smplcos ((COt t)::cos) = (COt(smplt t))::(smplcos cos)
      | smplcos ((COu u)::cos) = (COu(smplu u))::(smplcos cos)
in
    fun smplc (CT ct) =
	(case ct of
	     CTid(t1,t2) => CT(CTid(smplt t1,smplt t2))
	   | CTsim(t1,t2) => CT(CTsim(smplt t1,smplt t2))
	   | CTleq(t1,t2) => CT(CTleq(smplt t1,smplt t2))
	   | CTnoob t => CT(CTnoob(smplt t))
	   | CTrel t => CT(CTrel(smplt t))
	   | CTobcap(cos,ca) => CT(CTobcap(smplcos cos,ca))
	   | CTtime(tg,t) => CT(CTtime(tg,smplt t)))
      | smplc (CU cu) =
	     (case cu of
		  CUleq(u1,u2) => CU(CUleq(smplu u1,smplu u2))
		| CUnoob u => CU(CUnoob(smplu u))
		| CUobcap(us,ca) => CU(CUobcap(map smplu us,ca))
		| CUobct(u,cg) => CU(CUobct(smplu u,cg))
		| CUrel u => CU(CUrel(smplu u)))
      | smplc c = c   (* CA, CG, Cfalse *)
end

fun redobs [] = []
  | redobs ((COt Tbool)::cos) = (redobs cos)
  | redobs ((COt (Tchan(_,_,Uzero)))::cos) = redobs cos
  | redobs ((COt (Tchan(_,_,u)))::cos) = (COu u)::(redobs cos)
  | redobs (co::cos) = co::(redobs cos)

(* if 1st argument is true, returns free type varibale list,
     otherwise, returns free usage variable list. *)
local 
    fun remove [] s = []
      | remove (x::xs) s = if x=s then xs else x::(remove xs s)
in
    fun fvu typ (Uuof t) = fvt typ t
      | fvu typ (Uout(_,u)) = fvu typ u
      | fvu typ (Uin(_,u)) = fvu typ u
      | fvu typ (Upar(u1,u2)) = uniq ((fvu typ u1) @ (fvu typ u2))
      | fvu typ (Uglb(u1,u2)) = uniq ((fvu typ u1) @ (fvu typ u2))
      | fvu typ (Urec((Uvar s),u2)) = if typ then fvu typ u2
				      else remove (fvu typ u2) s 
      | fvu typ (Urep u) = fvu typ u
      | fvu _ (Urec _) = raise InternalError
      | fvu typ (Uvar s) = if typ then [] else [s]
      | fvu _ Uzero = []

    and fvt typ Tbool = []
      | fvt typ (Tchan(ts,_,u)) =
	uniq ((List.concat (map (fvt typ) ts)) @ (fvu typ u))
      | fvt typ (Tvar x) = if typ then [x] else []
      | fvt typ (Tadd(t1,t2)) = uniq ((fvt typ t1) @ (fvt typ t2))
      | fvt typ (Tglb(t1,t2)) = uniq ((fvt typ t1) @ (fvt typ t2))
      | fvt typ (Trep t1) = fvt typ t1
      | fvt typ (Tif t1) = fvt typ t1
end

fun fvco typ (COt t) = fvt typ t
  | fvco typ (COu u) = fvu typ u

fun fvcon typ (CT ct) =
    (case ct of
	 CTid(t1,t2) => uniq ((fvt typ t1) @ (fvt typ t2))
       | CTsim(t1,t2) => uniq ((fvt typ t1) @ (fvt typ t2))
       | CTleq(t1,t2) => uniq ((fvt typ t1) @ (fvt typ t2))
       | CTnoob(t1) => fvt typ t1
       | CTrel(t1) => fvt typ t1
       | CTobcap(cos,_) => uniq (List.concat (map (fvco typ) cos))
       | CTtime(_,t1) => fvt typ t1)
  | fvcon typ (CU cu) = 
	 (case cu of
	      CUleq(u1,u2) => uniq ((fvu typ u1) @ (fvu typ u2))
	    | CUnoob(u1) => fvu typ u1
	    | CUobcap(us,_) => uniq (List.concat (map (fvu typ) us))
	    | CUobct(u1,_) => fvu typ u1
	    | CUrel(u1) => fvu typ u1)
  | fvcon typ c = []
fun fvcons typ cs = uniq (List.concat (map (fvcon typ) cs))

(***************************************************************)
local
    fun zipid ([],[]) = []
      | zipid ([],_) = raise InternalError
      | zipid (_,[]) = raise InternalError
      | zipid (t1::ts1,t2::ts2) = CT(CTid(t1,t2))::(zipid(ts1,ts2))
in
    (* reduce constraints on types: phase 1.1
       reduce t1==t2's and t1~~t2's *)
    fun redct (cs1,cs2,CT(c)) =
	(case c of
	     CTid(Tvar s,t2) => 
		 if t2=(Tvar s) then
 		     (print ("From "^(ppct c)^": ");
		      print "it's true.\n";
		      (cs1,cs2,true))
		 else if member s (fvt true t2) then
 		     (print ("From "^(ppct c)^": ");
		      print "it contains recursive type.\n";
		      ([Cfalse],[],true))
		 else
		     (print ("From "^(ppct c)^": ");
		      print ("subst "^(ppt t2)^" for "^s^"\n");
		      (substcs(cs1,RT(s,t2)),substcs(cs2,RT(s,t2)),true))
	   | CTid(t1,Tvar s) => 
		 if member s (fvt true t1) then
		     (print ("From "^(ppct c)^": ");
		      print "it contains recursive type.\n";
		      ([Cfalse],[],true))
		 else
		     (print ("From "^(ppct c)^": ");
		      print ("subst "^(ppt t1)^" for "^s^"\n");
		      (substcs(cs1,RT(s,t1)),substcs(cs2,RT(s,t1)),true))
	   | CTid(Tbool,Tbool) =>
			 (print ("From "^(ppct c)^": it's true\n");
			  (cs1,cs2,true))
	   | CTid(Tbool,Tchan _) =>
			 (print ("From "^(ppct c)^": types don't match.\n");
			  ([Cfalse],[],true))
	   | CTid(Tchan(ts1,tg1,u1),Tchan(ts2,tg2,u2)) =>
			 if (len ts1) <> (len ts2) then raise InternalError
			 else 
			     let val newcs = zipid(ts1,ts2) in
				 (print ("From "^(ppct c)^": ");
				  print ("subst tag "^(Int.toString tg2)^" for tag "^(Int.toString tg1)^"\n");
				  print ("subst "^(ppu u2)^" for "^(ppu u1)^"\n");
				  print ("\""^(ppcons ", " newcs)^"\" added\n");
				  (
				   (substcs(substcs(cs1, RG(tg1,tg2)), RU(u1,u2)),
				    substcs(substcs(cs2 @ newcs,RG(tg1,tg2)), RU(u1,u2)),
				    true)))
			     end
	   | CTid(Tchan _,Tbool) =>
			     (print ("From "^(ppct c)^": types don't match.\n");
			      ([Cfalse],[],true))
			     
	   | CTsim(Tvar s,Tbool) => 
			     (print ("From "^(ppct c)^": ");
			      print ("subst Bool for "^s^"\n");
			      (substcs(cs1,RT(s,Tbool)), substcs(cs2,RT(s,Tbool)),true))
	   | CTsim(Tvar s,Tchan(ts,tg,u)) =>
			     if member s (List.concat (map (fvt true) ts)) then
				 (print ("From "^(ppct c)^": ");
				  print "it contains recursive type.\n";
				  ([Cfalse],[],true))
			     else
				 let val uu = Uvar(new "u") in
				     (print ("From "^(ppct c)^": ");
				      print ("subst "^(ppt (Tchan(ts,tg,uu)))^" for "^s^"\n");
				      (substcs(cs1,RT(s,Tchan(ts,tg,uu))),
				       substcs(cs2,RT(s,Tchan(ts,tg,uu))),true))
				 end
	   | CTsim(Tbool,Tvar(s)) => 
				 (print ("From "^(ppct c)^": ");
				  print ("subst Bool for "^s^"\n");
				  (substcs(cs1,RT(s,Tbool)), substcs(cs2,RT(s,Tbool)),true))
	   | CTsim(Tchan(ts,tg,u),Tvar(s)) =>
				 if member s (List.concat (map (fvt true) ts)) then
				     (print ("From "^(ppct c)^": ");
				      print "it contains recursive type.\n";
				      ([Cfalse],[],true))
				 else
				     let val uu = Uvar(new "u") in
					 (print ("From "^(ppct c)^": ");
					  print ("subst "^(ppt (Tchan(ts,tg,uu)))^" for "^s^"\n");
					  (substcs(cs1,RT(s,Tchan(ts,tg,uu))),
					   substcs(cs2,RT(s,Tchan(ts,tg,uu))),true))
				     end
	   | CTsim(Tbool,Tbool) => 
				     (print ("From "^(ppct c)^": it's true\n");
				      (cs1,cs2,true))
	   | CTsim(Tbool,Tchan _) =>
				     (print ("From "^(ppct c)^": types don't match.\n");
				      ([Cfalse],[],true))
	   | CTsim(Tchan(ts1,tg1,u1),Tchan(ts2,tg2,u2)) =>
				     if (len ts1) <> (len ts2) then 
					 (print ("From "^(ppct c)^": "^
						 "# of arguments don't match.\n");
					  ([Cfalse],[],true))
				     else 
					 let val newcs = zipid(ts1,ts2) in
					     (print ("From "^(ppct c)^": ");
					      print ("subst tag "^(Int.toString tg2)^" for tag "^
						     (Int.toString tg1)^"\n");
					      print ("\""^(ppcons ", " newcs)^"\" added\n");
					      (substcs(cs1, RG(tg1,tg2)),
					       substcs(cs2 @ newcs,RG(tg1,tg2)),true))
					 end
	   | CTsim(Tchan _,Tbool) => 
					 (print ("From "^(ppct c)^": types don't match.\n");
					  ([Cfalse],[],true))
	   | _ => ((CT c)::cs1, cs2, false)
					 )
      | redct (cs1, cs2, c) = (c::cs1,cs2,false)
end
(* redct *) 
	 
(* judges whether co is of the form
 'COu(u1) or COu(u2) or ... or COu(un) => ...' *)
local				       
    fun sub (us,[]) = SOME (rev us)
      | sub (_,(COt t)::cos) = NONE
      | sub (us,(COu u)::cos) = sub (u::us,cos)
in
    fun allcou cos = sub ([],cos)
end

(* phase 1.2 *)
fun redct2 (cs1,cs2,(CT c)) =
    (case c of
	 CTleq(Tvar s1,Tvar s2) =>
	     if s1=s2 then 
		 (print ("From "^(ppct c)^": it's true\n");
		  (cs1,cs2,true))
	     else ((CT c)::cs1,cs2,false)
       | CTleq(Tbool,Tbool) => 
		 (print ("From "^(ppct c)^": it's true\n");
		  (cs1,cs2,true))
       | CTleq(Tchan(_,_,u1),Tchan(_,_,u2)) =>
		 let val c1 = CU(CUleq(u1,u2)) in
		     (print ("From "^(ppct c)^": ");
		      print ((ppcon c1)^" added\n");
		      (cs1,cs2 @ [c1],true))
		 end
       | CTnoob(Tbool) =>
		 (print ("From "^(ppct c)^": it's true\n");
		  (cs1,cs2,true))
       | CTnoob(Tchan(_,_,u)) => 
		 let val c1 = CU(CUnoob(u)) in
		     (print ("From "^(ppct c)^": ");
		      print ((ppcon c1)^" added\n");
		      (cs1,cs2 @ [c1],true))
		 end
       | CTrel(Tbool) => 		 
		 (print ("From "^(ppct c)^": it's false.\n");
		  ([Cfalse],[],true))
       | CTrel(Tchan(_,_,u))=> 
		 let val c1 = CU(CUrel(u)) in
		     (print ("From "^(ppct c)^": ");
		      print ((ppcon c1)^" added\n");
		      (cs1,cs2 @ [c1],true))
		 end
       | CTobcap([],ca) => 	       
		 (print ("From "^(ppct c)^": it can be removed.\n");
		  (cs1,cs2,true))
       | CTobcap(cos,ca) =>
		 let val obs = redobs(cos) in
		     if (null obs) then
			 (print ("From: "^(ppct c)^": ");
			  print "reduced to true\n";
			  (cs1,cs2,true))
		     else if obs = cos then
			 ((CT c)::cs1,cs2,false)
			  else (case allcou obs of
				    NONE =>
					let val c1 = CT(CTobcap(obs,ca)) in
					    (print ("From: "^(ppct c)^": ");
					     print ("reduced to "^(ppcon c1)^"\n");
					     (c1::cs1,cs2,true))
					end
				  | SOME us =>
					let val c1 = CU(CUobcap(us,ca)) in
					    (print ("From: "^(ppct c)^": ");
					     print ("reduced to "^(ppcon c1)^"\n");
					     (c1::cs1,cs2,true))
					end)
		 end
       | CTtime(g,Tbool) =>
		 (print ("From "^(ppct c)^": it's true\n");
		  (cs1,cs2,true))
       | CTtime(g,Tchan(_,g1,u1)) =>
		 let val c1 = CU(CUobct(u1,CGleq(g,g1))) in
		     (print ("From "^(ppct c)^": ");
		      print ((ppcon c1)^" added\n");
		      (cs1,cs2 @ [c1],true))
		 end
       | _ => ((CT c)::cs1,cs2,false))
  (* option *)
  | redct2 (cs1,cs2,c as (CU(CUnoob u))) =
	 if u=Uzero then
	     (print ("From "^(ppcon c)^": it's true\n");
	      (cs1,cs2,true))
	 else
	     let val c1 = CU(CUleq(u,Uzero)) in
		 (print ("From "^(ppcon c)^": converted to "^(ppcon c1)^"\n");
		  (c1::cs1,cs2,true))
	     end
  | redct2 (cs1,cs2,c as (CU(CUobct(Uzero,_)))) =
	     (print ("From "^(ppcon c)^": it's true\n");
	      (cs1,cs2,true))
  | redct2 (cs1,cs2,c as (CU(CUleq(Uzero,_)))) =
	     (print ("From "^(ppcon c)^": it's true\n");
	      (cs1,cs2,true))
  | redct2 (cs1,cs2,c) = (c::cs1,cs2,false)

(* substitute []^g/0 for free type variables in constraints *)
local
    fun newchan v = RT(v,Tchan([],newtag(),Uzero))
	
    fun sub cs [] = cs
      | sub cs (rp::rps) = sub (substcs (cs,rp)) rps
in
    fun substftv cs =
	(print "phase1.2: substitution of free type variables.\n";
	 let val rps = map newchan (fvcons true cs) in
	     (print (pplist pprp "" ",\n" rps);
	      sub cs rps)
	 end)
end