(* reduce constraints: phase2 *)

(* constructs a disjunctive normal form *)
(* dnf is 
       ECAt(g1,g2)
     | ECor[ECand A1, ECand A2, ..., ECand An]
       where each Ai is a list of literal *)
local
    fun dnfsub_or [] = []
      | dnfsub_or ((ECor ecs1)::ecs2) = ecs1 @ (dnfsub_or ecs2)
      | dnfsub_or (ec::ecs) = ec::(dnfsub_or ecs)
    fun dnfsub_and [] = []
      | dnfsub_and ((ECand ecs1)::ecs2) = ecs1 @ (dnfsub_and ecs2)
      | dnfsub_and (ec::ecs) = ec::(dnfsub_and ecs)
    fun distrib [] = ECtrue
      | distrib [ec] = ec
      | distrib (ec::ecs) = 
	(case ec of
	     ECor ecs1 => 
		 let val (ECor ecs2) = distrib ecs in
		     ECor (List.concat 
			   (map (fn ec1 => map 
				 (fn ec2 => ECand (dnfsub_and [ec1,ec2]))
				 ecs2) ecs1))
		 end
	   | _ => let val (ECor ecs1) = (distrib ecs) in
		 ECor (map (fn ec1 => ECand (dnfsub_and [ec,ec1])) ecs1)
		  end)
in
    fun dnfeclr (ECor ecs) = ECor (dnfsub_or (map dnfeclr ecs))
      | dnfeclr (ECand ecs) = distrib (map dnfeclr ecs)
      | dnfeclr (ECAt(g1,g2)) = ECAt(g1,g2)
      | dnfeclr ec = ECor [ECand [ec]]
end
fun dnf (EC(ecl,ecr)) = EC(dnfeclr ecl,dnfeclr ecr)

(* simplify constraint: it works for dnf. *)
fun smpleclr (ECor ecs) =
    let val ecs = map smpleclr ecs in
	if member ECtrue ecs then ECtrue
	else (case (List.filter (fn x => x <> ECfalse) ecs) of
		  [] => ECfalse
		| ecs => ECor (uniq ecs))
    end
  | smpleclr (ECand ecs) =
    let val ecs = map smpleclr ecs in
	if member ECfalse ecs then ECfalse
	else (case (List.filter (fn x => x <> ECtrue) ecs) of
		  [] => ECtrue
		| ecs => ECand (uniq ecs))
    end
  | smpleclr ec = ec     
fun smplec (EC(ecl,ecr)) = EC(smpleclr ecl,smpleclr ecr)

(* misc. function *)
(* find(u,cs): find list of all constraints of form u<=U and the rest,
                where u is a usage variable (Uvar s).   *)
local
    fun findsub (_,[],culeqs,rest) = (rev culeqs,rev rest)
      | findsub (u,(c as (CU(CUleq(u1,u2))))::cs,culeqs,rest) =
	if u=u1 then
	    findsub(u,cs,c::culeqs,rest)
	else
	    findsub(u,cs,culeqs,c::rest)
      | findsub (u,c::cs,culeqs,rest) = findsub(u,cs,culeqs,c::rest)
in
    fun find u cs = findsub(u,cs,[],[])
end

(* get all 'u1<=U1, u1<=U2, ..." together into a single 'u1<=U' *)
fun makeglb v [] = CU(CUleq(Uvar v,Uvar v))
  | makeglb v [c] = c
  | makeglb v ((CU(CUleq(_,u1)))::(CU(CUleq(_,u2)))::cs) =
    makeglb v ((CU(CUleq(Uvar v,Uglb(u1,u2))))::cs)
  | makeglb _ _ = raise InternalError

(* 'findvs cs vs' extracts 'u1<=U1, u2<=U2, ..., un<=Un'
                           (ui = (Uvar vi), vi \in vs)
			   ('vs' are free usage variables)
                 from cs and returns the rest *)
local
    fun findvssub cs culeqs [] = (culeqs,cs)
      | findvssub cs culeqs (v::vs) =
	let val (cs1,rest) = find (Uvar v) cs in
	    findvssub rest ((makeglb v cs1)::culeqs) vs
	end
in
    fun findvs cs vs = findvssub cs [] vs
end

(* makes a normal solution "theta" from constraints
                     'u1<=U1, ..., u2<=U2' (each ui distinct) *)
fun normal [] theta = rev theta
  | normal (CU(CUleq(Uvar s,u))::cs) theta =
    if member s (fvu false u) then
	(* substitute recursive usage *)
	normal
	(substcs(cs,RU(Uvar s,Urec(Uvar s,u))))
	(RU(Uvar s,Urec(Uvar s,u))::theta)
    else
	(* substitute non-recursive usage *)
	normal 
	(substcs(cs,RU(Uvar s,u)))
	(RU(Uvar s,u)::theta)
  | normal _ _ = raise InternalError

fun grouping (obcaps,obcts,cas,rels) [] = (obcaps,obcts,cas,rels)
  | grouping (obcaps,obcts,cas,rels) ((CU(CUobcap(us,ca)))::cs)
    = grouping ((us,ca)::obcaps,obcts,cas,rels) cs
  | grouping (obcaps,obcts,cas,rels) ((CU(CUobct(u,cg)))::cs)
    = grouping (obcaps,(u,cg)::obcts,cas,rels) cs
  | grouping (obcaps,obcts,cas,rels) ((CU(CUrel u))::cs)
    = grouping (obcaps,obcts,cas,u::rels) cs
  | grouping (obcaps,obcts,cas,rels) ((CA(CAinc(a1,a2)))::cs)
    = grouping (obcaps,obcts,(CAinc(a1,a2))::cas,rels) cs
  | grouping _ _ = raise InternalError

(* substitution *)
fun subst u [] = u
  | subst u (rp::theta) = subst (substu (u,rp)) theta

fun redca (CAinc(a1,(Avar s))) =
    (case a1 of 
	 Az  =>  ECtrue
       | Ac  =>  ECAcap s
       | Ao  =>  ECAob s
       | Aco =>  ECand[ECAcap s,ECAob s]
       | _ => raise InternalError)

  | redca _ = raise InternalError

fun red_ca ca = 
    (println ("red_ca: processing "^(ppca ca));
     let val ec1 = EC(ECtrue, redca ca)
	 val ec2 = dnf ec1
	 val ec3 = smplec ec2
     in
	 (println ("result: "^(ppec ec1));
	  println ("dnf: "^(ppec ec2));
	  println ("smplified: "^(ppec ec3)^"\n");
	  ec3)
     end)
    
fun redcg (CGleq(g1,g2)) = ECAt(g1,g2)

(* reduce 'ob_i(U)' or 'ob_o(U)' for usages without 'rec' *)
fun red_u _ Uzero = ECfalse
  | red_u input (Uout(a,u)) = if input then ECfalse 
			    else redca (CAinc(Ao,a))
  | red_u input (Uin(a,u)) = if input then redca (CAinc(Ao,a))
			   else ECfalse 
  | red_u input (Upar(u,v)) = ECor[red_u input u,red_u input v]
  | red_u input (Uglb(u,v)) = ECand[red_u input u,red_u input v]
  | red_u input (Uvar s) = if input then (ECAobi s) else (ECAobo s)
  | red_u input (Urep u) = red_u input u
  | red_u _ _ = raise InternalError    (* Urec, Uuof *)

fun redu u =
    let val ec1 = ECor [red_u true u,red_u false u] in
	(println ("\tredu: processing ob("^(ppu u)^")");
	 println ("\tresult: "^(ppeclr ec1));
	 ec1)
    end

local
    (* reduce 'mayob(U)' *)
    fun mayob _ Uzero = ECfalse
      | mayob _ (Uvar s) = ECfalse
      | mayob inp (Uout(a,u)) = if inp then ECfalse else redca(CAinc(Ao,a))
      | mayob inp (Uin(a,u)) = if inp then redca(CAinc(Ao,a)) else ECfalse
      | mayob inp (Upar(u,v)) = ECor[mayob inp u, mayob inp v]
      | mayob inp (Uglb(u,v)) = ECor[mayob inp u, mayob inp v]
      | mayob inp (Urec(u,v)) = mayob inp v
      | mayob inp (Urep u) = mayob inp u 
      | mayob _ _ = raise InternalError      (* Uuof *)
in
    (* reduce 'ob^*_i(U,theta)' or 'ob^*_o(U,theta)' for closed usage U *)
    fun reducl inp Uzero _ = ECfalse
      | reducl inp (Uvar s) _ = ECtrue
      | reducl inp (Uout(a,u)) _ = if inp then ECfalse else redca(CAinc(Ao,a))
      | reducl inp (Uin(a,u)) _ = if inp then redca(CAinc(Ao,a)) else ECfalse
      | reducl inp (Upar(u,v)) theta = ECor[reducl inp u theta, reducl inp v theta]
      | reducl inp (Uglb(u,v)) theta = ECand[reducl inp u theta, reducl inp v theta]
      | reducl inp (Urec(u,v)) theta = 
	let val u1 = subst v theta in
	    ECand[mayob inp u1, reducl inp u (theta @ [RU(u,Urec(u,u1))])]
	end
      | reducl inp (Urep u) theta = reducl inp u theta
      | reducl _ _ _ = raise InternalError    (* Uuof *)
end

(* red_u for closed usages *)
fun red_u_cl input u =
    let val ec1 = ECor [red_u true u,red_u false u] in
	(println ("\tredu: processing ob("^(ppu u)^")");
	 println ("\tresult: "^(ppeclr ec1));
	 ec1)
    end

fun red_obcap (us,ca) = 
    (println ("red_obcap: processing "^(ppcu (CUobcap(us,ca))));
     let val ec1 = EC(ECor(map redu us),redca ca)
	 val ec2 = dnf ec1
	 val ec3 = smplec ec2 in
	     (println ("result: "^(ppec ec1));
	      println ("dnf: "^(ppec ec2));
	      println ("smplified: "^(ppec ec3)^"\n");
	      ec3)
     end)

fun red_obct(u,ct) = 
    (println ("red_obct: processing "^(ppcu (CUobct(u,ct))));
     let val ec1 = EC(redu u,redcg ct)
	 val ec2 = dnf ec1
	 val ec3 = smplec ec2
     in
	 (println ("result: "^(ppec ec1));
	  println ("dnf: "^(ppec ec2));
	  println ("smplified: "^(ppec ec3)^"\n");
	  ec3)
     end)

fun comp Uzero = []
  | comp (Upar(u,v)) = uniq ((comp u) @ (comp v))
  | comp u = [u]

fun subuexp Uzero = []
  | subuexp (Uvar s) = [Uvar s]
  | subuexp (Uin(a,u)) = uniq ([Uin(a,u)] @ (subuexp u))
  | subuexp (Uout(a,u)) = uniq ([Uout(a,u)] @ (subuexp u))
  | subuexp (Upar(u,v)) = uniq ((subuexp u) @ (subuexp v))
  | subuexp (Uglb(u,v)) = uniq ([Uglb(u,v)] @ (subuexp u) @ (subuexp v))
  | subuexp (Urec(u,v)) = 
    uniq ([Urec(u,v)] @ 
	  (map (fn u0 => substu(u0,RU(u,Urec(u,v)))) (subuexp v)))
  | subuexp (Urep u) = uniq ((Urep u) :: (subuexp u))
  | subuexp _ = raise InternalError

fun cano basis Uzero = times (len basis) 0
  | cano basis (Upar(u,v)) = vecsum (cano basis u) (cano basis v)
  | cano (b::basis) u = if b=u then 1::(times (len basis) 0)
			else 0::(cano basis u)
  | cano [] u = raise InternalError

fun judge1 (Uout _) = true
  | judge1 _ = false

local
    fun find u [] = raise InternalError
      | find u ((CU(CUleq(u1,U)))::culeqs) = 
	if u=u1 then U 
	else find u culeqs
      | find _ _ = raise InternalError

    fun uvecsub culeqs basis (Uglb(u,v)) = 
	[vecdiff (cano basis u) (cano basis (Uglb(u,v))),
	 vecdiff (cano basis v) (cano basis (Uglb(u,v)))]
      | uvecsub culeqs basis (Uvar s) =
	[vecdiff (cano basis (find (Uvar s) culeqs)) (cano basis (Uvar s))]
      | uvecsub culeqs basis (Urec(u,v)) =
	[vecdiff (cano basis (substu(v,RU(u,Urec(u,v))))) (cano basis (Urec(u,v)))]
      | uvecsub culeqs basis (Urep u) =
	[cano basis u]
      | uvecsub culeqs basis (Uin(a,u)) =
	let
	    fun uvecsub2 (Uin(a,u)) (Uout(b,v)) =
		vecdiff (vecdiff (vecsum (cano basis u) (cano basis v))
			 (cano basis (Uin(a,u))))
		(cano basis (Uout(b,v)))
	      | uvecsub2 _ _ = raise InternalError
	in
	    map (uvecsub2 (Uin(a,u))) (List.filter judge1 basis)
	end
      | uvecsub _ _ _ = []
in
    fun uvec culeqs basis = uniq (List.concat (map (uvecsub culeqs basis) basis))
end

(* vs means a set of atomic usages in basis *)
local
    fun vec [] _ = []
      | vec (b::basis) vs = (if member b vs then 1 else 0)::(vec basis vs)
    fun makeslack [] = []
      | makeslack (v::vs) = 
	if v=1 then 
	    ((~1)::(times (len vs) 0))::(map (secl 0 op::) (makeslack vs))
	else
	    map (secl 0 op::) (makeslack vs)
    fun transp ([]::_) = []
      |  transp mat = (map hd mat)::(transp (map tl mat))
in
    fun makeeqns x basis uvecs vs =
	let 
	    val z = vec basis vs
	in
	    transp ((vecdiff z x)::uvecs @ (makeslack z))
	end
end

fun solvable mat = msolve mat = 0

fun deriv f culeqs u =
    let val basis = 
	uniq (List.concat (map subuexp 
			   ([u] @ (map
				   (fn (CU(CUleq(u1,u2))) => u2
				     | _ => raise InternalError) culeqs))))
	val uvecs = uvec culeqs basis
	val x = cano basis u
	val cand = powerset basis
	fun derivsub oks [] = oks
	  | derivsub oks (vs::cand) =
	    (print ("\tderivsub: checking {"^(ppus ", " vs)^"} - ");
	     if solvable (makeeqns x basis uvecs vs) then
		 (println "reachable.";
		  derivsub (vs::oks) 
		  (lremove cand (map (uniq o List.concat) (powerset (vs::oks)))))
	     else
		 (println "not reachable.";
		  derivsub oks cand))
    in
	(println ("deriv:\tu = "^(ppu u)^"\n\tbasis: {"^(ppus ", " basis)^"}"^
		  "\n\tculeqs: "^(ppcons "\n" culeqs)); 
	 List.concat (map f (derivsub [] cand)))
    end

(* us is a list of atomic usages *)
(* con1 requires backtrack *)
local
    fun con1sub us au =
	let val rest = remove us [au] in
	    (case au of
		 Uin(Avar s,u) => [EC(ECAcap s,ECor (map (red_u false) rest))]
	       | Uout(Avar s,u) => [EC(ECAcap s,ECor (map (red_u true) rest))]
	       | _ => [])
	end
in
    fun con1 us = 
	let val ecs = List.concat (map (con1sub us) us) in
	    (println ("\tcon1: usage: "^(ppus ", " us)^
		      "\n\t\tresult:\n\t\t"^(pplist ppec "" "\n\t\t" ecs));
	     map (smplec o dnf) ecs)	
	end
end

local
fun mustob _ Uzero = ECfalse
  | mustob input (Uout(a,u)) = if input then ECtrue
			    else redca (CAinc(Ao,a))
  | mustob input (Uin(a,u)) = if input then redca (CAinc(Ao,a))
			   else ECtrue
  | mustob input (Upar(u,v)) = ECand[mustob input u,mustob input v]
  | mustob input (Uglb(u,v)) = ECand[mustob input u,mustob input v]
  | mustob input (Uvar s) = if input then (ECAobi s) else (ECAobo s)
  | mustob input (Urep u) = mustob input u
  | mustob _ _ = raise InternalError    (* Urec, Uuof *)
    fun con2sub us au =
	let val rest = remove us [au] in
	    (case au of
		 Uin(Avar s,u) => [EC(ECAcap s,ECand (map (mustob false) rest))]
	       | Uout(Avar s,u) => [EC(ECAcap s,ECand (map (mustob true) rest))]
	       | _ => [])
	end
in
    fun con2 us = 
	let val ecs = List.concat (map (con2sub us) us) in
	    (println ("\tcon2: usage: "^(ppus ", " us)^
		      "\n\t\tresult:\n\t\t"^(pplist ppec "" "\n\t\t" ecs));
	     map (smplec o dnf) ecs)	
	end
end

fun red_rel con_func cs u = 
    (println ("red_rel: processing "^(ppcu (CUrel u)));
     let val ecs = deriv con_func cs u in
	 (println ("result:\n"^(ppecs ecs)^"\n");
	  ecs)
     end)
    
fun makeextra theta v =
    let val u = subst (Uvar v) theta
	val ecs = [EC(ECAobi v,reducl true u []),
		   EC(ECAobo v,reducl false u [])]
	val ecs2 = map (smplec o dnf) ecs in
	    (println ("extra condition:\n\tob_i("^v^
		      ") -> ob_i("^(ppu u)^"),\n"^
		      "\tob_o("^v^") -> ob_o("^(ppu u)^") added.");
	     println ("\tresult:\n\t"^(pplist ppec "" "\n\t" ecs));
	     println ("\tsmplified:\n\t"^(pplist ppec "" "\n\t" ecs2)^"\n");
	     ecs2)
    end

(* main function *)

fun redcu con_func cs =
    let 
	val fvs = fvcons false cs
	val (culeqs,rest) = findvs cs fvs
	val dummy = println ("phase 2: reduce constraints on usages\nconstraints:\n"^(ppcons "\n" (culeqs@rest))^"\n")
	val (obcaps,obcts,cas,rels) = grouping ([],[],[],[]) rest
	val theta = normal culeqs []
	val ecs0 = List.concat (map (makeextra theta) fvs)  
	val ecs1 = map red_obcap obcaps
	val ecs2 = map red_obct obcts
	val ecs3 = uniq (map red_ca cas)
	val ecs4 = List.concat (map (red_rel con_func culeqs) rels)
	val ecs5 = ecs0 @ ecs1 @ ecs2 @ ecs3 @ ecs4
	val ecs6 = let
		       fun judge (EC(ECtrue,ECtrue)) = false
			 | judge (EC(ECfalse,_)) = false
			 | judge _ = true
		   in
		       List.filter judge ecs5
		   end
    in
	(println ("phase 2 result:\n"^(ppecs ecs5)^"\n");
	 println ("simplified:\n"^(ppecs ecs6)^"\n");
	 ecs6)
    end


