(* cycle detection *)
fun nexts i graph =
    List.concat (map (fn (m,n) => if m=i then [n] else []) graph)

fun pppath [x,y] = ppcg (CGleq(x,y)) 
  | pppath (x::y::z::xs) = (ppcg (CGleq(x,y)))^", " ^ (pppath (y::z::xs))
  | pppath _ = raise InternalError

fun pathsort graph = 
    let
	fun sort ([],_,visited) = visited
	  | sort (x::xs,path,visited) =
	    if member x path then
		(println ("NG - "^(pppath (rev (x::path)))^" contains a cycle.\n");
		 raise InternalError)
	    else
		sort (xs, path, 
		      if member x visited then visited
		      else x::sort(nexts x graph,x::path,visited))
	val (starts,_) = ListPair.unzip graph
    in
	(println "time tag order checking.";
	 println ("OK - time tag order: "^
		  (pplist ppg "" " < " (sort(starts,[],[])))^"\n");
	 true)
	handle InternalError => false
    end 

(* 'atms' are variables assigned "true", provided dnf *) 
fun evalsub atms ECtrue = ECtrue
  | evalsub atms ECfalse = ECfalse
  | evalsub atms (ECand eclrs) = smpleclr (ECand (map (evalsub atms) eclrs))
  | evalsub atms (ECor eclrs) = smpleclr (ECor (map (evalsub atms) eclrs))
  | evalsub atms ecatm = if member ecatm atms then ECtrue else ecatm
fun eval atms (EC(ecl,ecr)) = EC(evalsub atms ecl, evalsub atms ecr)

(* 'ecrs' are constraints that we want to satisfy *)
local
    fun satsub1 atms ecs [] cgs = if (null cgs) then true else pathsort cgs
      | satsub1 atms ecs (ecr::ecrs) cgs =
	(case ecr of
	     ECtrue => satsub1 atms ecs ecrs cgs
	   | ECfalse => false
	   | (ECAt (t1,t2)) => satsub1 atms ecs ecrs ((t1,t2)::cgs)
	   | (ECand eclrs) =>
		 (println ("constraints {"^
			   (pplist ppeclr "" ", " eclrs)^
			   "} added.\n");
		  satmain1 (atms@eclrs) ecs)
	   | (ECor ((ECand ecand)::eclrs)) => 
		 if (satsub1 atms ecs ((ECand ecand)::ecrs) cgs) then
		     true
		 else
		     if len eclrs > 0 then
			 (println "backtrack!";
			  satsub1 atms ecs [ECor eclrs] cgs)
		     else false
	   | (ECor (_::_)) => raise InternalError
	   | (ECor []) => false
	   | _ => raise InternalError)

    and satsub2 atms ecs [] cgs = if (null cgs) then true else pathsort cgs
      | satsub2 atms ecs (ecr::ecrs) cgs =
	(case ecr of
	     ECtrue => satsub2 atms ecs ecrs cgs
	   | ECfalse => false
	   | (ECAt (t1,t2)) => satsub2 atms ecs ecrs ((t1,t2)::cgs)
	   | (ECand eclrs) =>
		 (println ("constraints {"^
			   (pplist ppeclr "" ", " eclrs)^
			   "} added.\n");
		  satmain2 (atms@eclrs) ecs)
	   | (ECor [ECand ecand]) =>
		 satsub2 atms ecs ((ECand ecand)::ecrs) cgs
	   | (ECor []) => false
	   | (ECor (_::_)) => raise InternalError
	   | _ => raise InternalError)
	     
    and satmain1 atms ecs = 
	let 
	    val ecs = uniq ecs
	    val ecs1 = map (eval atms) ecs
	    val ecs2 = List.filter (fn (EC(ecl,_)) => ecl=ECtrue) ecs1
	in
	    (println ("sat: {"^(pplist ppeclr "" ", " atms)^"} is true.");
	     println ("constraints are:\n"^(ppecs ecs)^"\n");
	     println ("evaluated:\n"^(ppecs ecs1)^"\n");
	     if null ecs2 then true
	     else
		 satsub1 atms ecs1 (map (fn (EC(l,r)) => r) ecs2) [])
	end
    and satmain2 atms ecs = 
	let 
	    val ecs = uniq ecs
	    val ecs1 = map (eval atms) ecs
	    val ecs2 = List.filter (fn (EC(ecl,_)) => ecl=ECtrue) ecs1
	in
	    (println ("sat: {"^(pplist ppeclr "" ", " atms)^"} is true.");
	     println ("constraints are:\n"^(ppecs ecs)^"\n");
	     println ("evaluated:\n"^(ppecs ecs1)^"\n");
	     if null ecs2 then true
	     else
		 satsub2 atms ecs1 (map (fn (EC(l,r)) => r) ecs2) [])
	end
in
    fun sat f ecs = 
	(println "phase3: satisfiability of extended constraints.";
	 if (f [] ecs) then
	     (println "constraints are satisfiable."; true)
	 else (println "constraints are not satisfiable."; false))
    val sat1 = sat satmain1
    val sat2 = sat satmain2
end