(* simplex method for minimum value *)
local 
    open Int
in
    fun gcd m n = if m=0 then n
	      else gcd (n mod m) m
    fun norm (x,y) =
	let val plus = (x>=0 andalso y>0) orelse (x<0 andalso y<0)
	    val x = abs x
	    val y = abs y
	    val g = gcd x y
	in
	    if plus then (x div g,y div g)
	    else (~(x div g),y div g)
	end
    fun sum (x,y) (z,w) = norm (x*w+z*y,y*w)
    fun prod (x,y) (z,w) = norm (x*z,y*w)
    fun diff (x,y) (z,w) = norm (x*w-z*y,y*w)
    fun quo a b = norm (a,b)
    fun toratio x = (x,1)
    fun pos (x,y) = x>0
    fun neg (x,y) = x<0
    fun zero (x,y) = x=0
    fun floor (x,y) = x div y 
end 
	
datatype pnum =
    NegInf
  | PosInf
  | Num of int
fun less NegInf _ = true
  | less _ NegInf = false
  | less _ PosInf = true
  | less PosInf _ = false
  | less (Num m) (Num n) = m < n

(* for rational number *)
datatype rpnum =
    RNegInf
  | RPosInf
  | RNum of int * int

(* diciedes a pivot row on the column 'col' *)
local
    fun rless RNegInf _ = true
      | rless _ RNegInf = false
      | rless _ RPosInf = true
      | rless RPosInf _ = false
      | rless (RNum m) (RNum n) = neg (diff m n)
    fun req RNegInf RNegInf = true
      | req RPosInf RPosInf = true
      | req (RNum m) (RNum n) = zero (diff m n)
      | req _ _ = false
    fun find _ _ [] = raise InternalError
      | find _ [] _ = raise InternalError
      | find pos (obj::objs) (x::xs) =
	if x=1 then if obj=0 then pos
		    else find (pos+1) objs xs
	else find (pos+1) objs xs
	    
    fun piv _ (RNum minv,index,minl) _ [] = (minv, minl)
      | piv _ _ _ [] = raise InternalError
      | piv _ _ [] _ = raise InternalError
      | piv col (minv,index,minl) (i::ind) ((x::xs)::rows) = 
	let val colv = List.nth(xs,col) in
	    if colv <= 0 then
		piv col (minv,index,minl) ind rows
	    else
		let val ratio = quo x colv in
		    if rless (RNum ratio) minv then
			piv col (RNum ratio,i,x::xs) ind rows
		    else 
			if req (RNum ratio) minv then
			    if i < index then
				piv col (minv,i,x::xs) ind rows
			    else
				piv col (minv,index,minl) ind rows
			else
			    piv col (minv,index,minl) ind rows
		end
	end
      | piv _ _ _ ([]::_) = raise InternalError
in
    fun pivotrow col index rows = piv col (RPosInf,~1,[]) index rows
end

fun scalarprod k xs = map (secl k op* ) xs
fun vecsum xs ys = ListPair.map op+ (xs,ys)
fun vecdiff xs ys = vecsum xs (scalarprod (~1) ys)

fun pivot _ _ [] = []
  | pivot col minl (row::rows) = 
    (vecsum 
     (scalarprod (~(List.nth(row,col+1))) minl)
     row) ::
    (pivot col minl rows)
    
(* performs one iteration: 'index' is column indices of basis *)
local
    fun first col (x::xs) = if x<0 then col else first (col+1) xs 
      | first _ [] = ~1 
    fun firstc xs = first 0 xs
in
    fun iter (_,[]) = raise InternalError
      | iter (index,row::rows) = 
	let 
	    val col = firstc (tl row)
	in
	    if col<0 then ([],[])
	    else
		let
		    val (minv,minl) = pivotrow col index rows
		    val a = List.nth(minl,col+1)
		    (* makes a new constraint equation:
		     'a' is a demoninator *)
		    fun makenew _ [] = [1]   (* slack var. *)
		      | makenew pos (x::xs) = 
			(if member pos index then 0 else floor (quo x a))::
			     (makenew (pos+1) xs)
		    val newminl = (floor (quo (hd minl) a)) :: 
			(makenew 0 (tl minl))
		in
		    (index @ [col],
		     pivot col newminl (map (fn xs => xs @ [0]) (row::rows))
		     @ [newminl])
		    
		end
	end
end

(* returns a minimum value *)
fun mini (index,rows) = 
    let val (newind,newrows) = iter (index,rows) in
	if null newrows then ~(hd(hd rows)) else mini (newind,newrows)
    end

local
    (* identity matrix of dim. n *)
    fun idmat 1 = [[1]]
      | idmat n = (1::(times (n-1) 0))::(map (secl 0 op::) (idmat (n-1)))
	
    fun upto 0 = [0]
      | upto n = (upto (n-1)) @ [n]

    fun fold _ [] = raise InternalError
      | fold _ [x] = x
      | fold f (x::y::xs) = fold f ((f x y)::xs)
in
    (* adds slack variables and objective function *)
    fun makebasis rows = 
	let val l = length rows
	    val sums = scalarprod ~1 (fold vecsum rows)
	in
	    (map (secl (len (hd rows)-1) op+) (upto (l-1)),
	     (sums @ (times l 0))::(ListPair.map op@ (rows,idmat l)))
	end
end

(* make constants be non-negative *)
fun msolve rows = 
    mini (makebasis (map (fn x=> if hd x >= 0 then x else scalarprod (~1) x) rows))
	    
